@@ -1,6 +1,6 @@
 // device_pmem.c : implementation of the rekall winpmem memory acquisition device.
 //
-// (c) Ulf Frisk, 2018-2022
+// (c) Ulf Frisk, 2018-2025
 // Author: Ulf Frisk, pcileech@frizk.net
 //
 #include "leechcore.h"
@@ -183,7 +183,7 @@ BOOL DevicePMEM_SvcStart(_In_ PLC_CONTEXT ctxLC)
                 pDriverFile = NULL;
                 break;
             }
-            ZeroMemory(szDriverFile, _countof(szDriverFile));
+            strcpy_s(szDriverFile, _countof(szDriverFile), ctxLC->Config.szDevice);
         }
     }
     if(fopen_s(&pDriverFile, szDriverFile, "rb") || !pDriverFile) {
@@ -387,7 +387,7 @@ BOOL DevicePMEM_Open(_Inout_ PLC_CONTEXT ctxLC, _Out_opt_ PPLC_CONFIG_ERRORINFO
 }
 
 #endif /* _WIN32 */
-#ifdef LINUX
+#if defined(LINUX) || defined(MACOS)
 
 _Success_(return)
 BOOL DevicePMEM_Open(_Inout_ PLC_CONTEXT ctxLC, _Out_opt_ PPLC_CONFIG_ERRORINFO ppLcCreateErrorInfo)
@@ -397,4 +397,4 @@ BOOL DevicePMEM_Open(_Inout_ PLC_CONTEXT ctxLC, _Out_opt_ PPLC_CONFIG_ERRORINFO
     return FALSE;
 }
 
-#endif /* LINUX */
+#endif /* LINUX || MACOS */
