@@ -2,7 +2,7 @@
 // 
 // tested with vmware workstation 16.1.2
 //
-// (c) Ulf Frisk, 2022
+// (c) Ulf Frisk, 2022-2025
 // Author: Ulf Frisk, pcileech@frizk.net
 //
 #include "leechcore.h"
@@ -60,7 +60,7 @@ BOOL DeviceVMWare_Open_GetRange(_In_ DWORD dwPID, _Out_writes_opt_(42) LPWSTR ws
 {
     DWORD cch;
     HANDLE hProcess = 0;
-    QWORD va = 0x0000001000000000;
+    QWORD va = 0x0000000000000000;
     WCHAR wsz[MAX_PATH + 1] = { 0 };
     MEMORY_BASIC_INFORMATION BI;
     hProcess = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, dwPID);
@@ -203,7 +203,7 @@ BOOL DeviceVMWare_Open(_Inout_ PLC_CONTEXT ctxLC, _Out_opt_ PPLC_CONFIG_ERRORINF
 }
 
 #endif /* _WIN32 */
-#ifdef LINUX
+#if defined(LINUX) || defined(MACOS)
 
 _Success_(return)
 BOOL DeviceVMWare_Open(_Inout_ PLC_CONTEXT ctxLC, _Out_opt_ PPLC_CONFIG_ERRORINFO ppLcCreateErrorInfo)
@@ -213,4 +213,4 @@ BOOL DeviceVMWare_Open(_Inout_ PLC_CONTEXT ctxLC, _Out_opt_ PPLC_CONFIG_ERRORINF
     return FALSE;
 }
 
-#endif /* LINUX */
+#endif /* LINUX || MACOS */
