@@ -1,6 +1,6 @@
 // leechrpcclient.c : implementation of the remote procedure call (RPC) client.
 //
-// (c) Ulf Frisk, 2018-2022
+// (c) Ulf Frisk, 2018-2025
 // Author: Ulf Frisk, pcileech@frizk.net
 //
 #include "leechcore.h"
@@ -8,69 +8,30 @@
 #include "leechcore_internal.h"
 #include "leechrpc.h"
 #include "util.h"
+#include "oscompatibility.h"
+#include <leechgrpc.h>
 
 #ifdef _WIN32
+
 #include <rpc.h>
+#include <ntsecapi.h>
+#include <wincred.h>
 #include "leechrpc_h.h"
 
+#endif /* _WIN32 */
+#if defined(LINUX) || defined(MACOS)
+
+#define LeechRPC_CompressClose(ctxCompress)
+#define LeechRPC_CompressInitialize(ctxCompress)                    FALSE
+#define LeechRPC_Compress(ctxCompress, pMsg, fCompressDisable)
+#define LeechRPC_Decompress(ctxCompress, pMsgIn, ppMsgOut)          FALSE
+
+#endif /* LINUX || MACOS */
+
 //-----------------------------------------------------------------------------
 // CORE FUNCTIONALITY BELOW:
 //-----------------------------------------------------------------------------
 
-_Success_(return)
-BOOL Util_GetBytesPipe(_In_ HANDLE hPipe_Rd, _Out_writes_opt_(cb) PBYTE pb, _In_ DWORD cb)
-{
-    DWORD cbReadTotal = 0, cbRead = 0;
-    while((cbReadTotal < cb) && ReadFile(hPipe_Rd, pb + cbReadTotal, cb - cbReadTotal, &cbRead, NULL)) {
-        cbReadTotal += cbRead;
-    }
-    return (cb == cbReadTotal);
-}
-
-/*
-* Send and receive data to the pipe server. First write the pMsgIn then listen
-* for a pMsgOut response from the server and read it.
-* NB! USER_FREE: *ppMsgOut
-* -- ctx
-* -- pMsgIn
-* -- pcbMsgOut
-* -- ppMsgOut
-* -- return
-*/
-_Success_(return)
-BOOL LeechRPC_SubmitCommand_Pipe(_In_ PLEECHRPC_CLIENT_CONTEXT ctx, _In_ PLEECHRPC_MSG_HDR pMsgIn, _Out_ PDWORD pcbMsgOut, _Out_ PPLEECHRPC_MSG_HDR ppMsgOut)
-{
-    static SRWLOCK LockSRW = SRWLOCK_INIT;
-    DWORD cbWrite = 0;
-    LEECHRPC_MSG_HDR Hdr = { 0 };
-    PLEECHRPC_MSG_HDR pMsgOut;
-    AcquireSRWLockExclusive(&LockSRW);
-    if(!pcbMsgOut || !ppMsgOut) { goto fail; }
-    // 1: write contents to pipe
-    if(!ctx->hPipeMem_Wr) { goto fail; }
-    if(!WriteFile(ctx->hPipeMem_Wr, (PVOID)pMsgIn, pMsgIn->cbMsg, &cbWrite, NULL)) { goto fail; }
-    // 2: read resulting contents : header
-    if(!Util_GetBytesPipe(ctx->hPipeMem_Rd, (PBYTE)&Hdr, sizeof(LEECHRPC_MSG_HDR))) { goto fail; }
-    if((Hdr.dwMagic != LEECHRPC_MSGMAGIC) || (Hdr.cbMsg > 0x04000000)) { goto fail; }
-    pMsgOut = (PLEECHRPC_MSG_HDR)LocalAlloc(0, Hdr.cbMsg);
-    if(!pMsgOut) { goto fail; }
-    memcpy(pMsgOut, &Hdr, sizeof(LEECHRPC_MSG_HDR));
-    // 3: read resulting contents : data
-    if(pMsgOut->cbMsg > sizeof(LEECHRPC_MSG_HDR)) {
-        if(!Util_GetBytesPipe(ctx->hPipeMem_Rd, ((PBYTE)pMsgOut) + sizeof(LEECHRPC_MSG_HDR), pMsgOut->cbMsg - sizeof(LEECHRPC_MSG_HDR))) {
-            LocalFree(pMsgOut);
-            goto fail;
-        }
-    }
-    ReleaseSRWLockExclusive(&LockSRW);
-    *pcbMsgOut = pMsgOut->cbMsg;
-    *ppMsgOut = pMsgOut;
-    return TRUE;
-fail:
-    ReleaseSRWLockExclusive(&LockSRW);
-    return FALSE;
-}
-
 _Success_(return)
 BOOL LeechRPC_SubmitCommand(_In_ PLC_CONTEXT ctxLC, _In_ PLEECHRPC_MSG_HDR pMsgIn, _In_ LEECHRPC_MSGTYPE tpMsgRsp, _Out_ PPLEECHRPC_MSG_HDR ppMsgOut)
 {
@@ -78,6 +39,7 @@ BOOL LeechRPC_SubmitCommand(_In_ PLC_CONTEXT ctxLC, _In_ PLEECHRPC_MSG_HDR pMsgI
     error_status_t error;
     BOOL fOK;
     DWORD cbMsgOut = 0;
+    SIZE_T cbMsgOutSize = 0;
     PLEECHRPC_MSG_BIN pMsgOutDecompress = NULL;
     // fill out message header given a message type
     pMsgIn->dwMagic = LEECHRPC_MSGMAGIC;
@@ -104,27 +66,32 @@ BOOL LeechRPC_SubmitCommand(_In_ PLC_CONTEXT ctxLC, _In_ PLEECHRPC_MSG_HDR pMsgI
         default:
             return FALSE;
     }
-    // submit message to RPC server or PIPE parent process.
-    if(ctx->fIsRpc) {
-        // RPC connection method:
+    // submit message to RPC server:
+    *ppMsgOut = NULL;
+    pMsgIn->dwRpcClientID = ctxLC->Rpc.dwRpcClientId;
+    pMsgIn->flags = ctxLC->Rpc.fCompress ? 0 : LEECHRPC_FLAG_NOCOMPRESS;
+    if(ctx->fIsProtoRpc || ctx->fIsProtoSmb) {
+        // RPC (over tcp or smb) connection methods:
+        error = E_FAIL;
+#ifdef _WIN32
         __try {
-            pMsgIn->dwRpcClientID = ctxLC->Rpc.dwRpcClientId;
-            pMsgIn->flags = ctxLC->Rpc.fCompress ? 0 : LEECHRPC_FLAG_NOCOMPRESS;
             error = LeechRpc_ReservedSubmitCommand(ctx->hRPC, pMsgIn->cbMsg, (PBYTE)pMsgIn, &cbMsgOut, (PBYTE*)ppMsgOut);
         } __except(EXCEPTION_EXECUTE_HANDLER) { error = E_FAIL; }
+#endif /* _WIN32 */
         if(error) {
-           *ppMsgOut = NULL;
-            return FALSE;
-        }
-    } else {
-        // PIPE connection method:
-        pMsgIn->dwRpcClientID = ctxLC->Rpc.dwRpcClientId;
-        pMsgIn->flags = LEECHRPC_FLAG_NOCOMPRESS;
-        if(!LeechRPC_SubmitCommand_Pipe(ctx, pMsgIn, &cbMsgOut, ppMsgOut)) {
             *ppMsgOut = NULL;
             return FALSE;
         }
     }
+    else if(ctx->fIsProtoGRpc) {
+        // gRPC (over tcp) connection method:
+        fOK = ctx->grpc.hGRPC && ctx->grpc.pfn_leechgrpc_client_submit_command(ctx->grpc.hGRPC, (PBYTE)pMsgIn, pMsgIn->cbMsg, (PBYTE*)ppMsgOut, &cbMsgOutSize);
+        if(!fOK) {
+            *ppMsgOut = NULL;
+            return FALSE;
+        }
+        cbMsgOut = (DWORD)cbMsgOutSize;
+    }
     // sanity check non-trusted incoming message from RPC server.
     fOK = (cbMsgOut >= sizeof(LEECHRPC_MSG_HDR)) && *ppMsgOut && ((*ppMsgOut)->dwMagic == LEECHRPC_MSGMAGIC);
     fOK = fOK && ((*ppMsgOut)->tpMsg <= LEECHRPC_MSGTYPE_MAX) && ((*ppMsgOut)->cbMsg == cbMsgOut) && (cbMsgOut < 0x10000000);
@@ -214,6 +181,17 @@ VOID LeechRPC_KeepaliveThreadClient(_In_ PLC_CONTEXT ctxLC)
 
 VOID LeechRPC_RpcClose(PLEECHRPC_CLIENT_CONTEXT ctx)
 {
+    // Close the gRPC connection:
+    if(ctx->grpc.hGRPC) {
+        ctx->grpc.pfn_leechgrpc_client_free(ctx->grpc.hGRPC);
+        ctx->grpc.hGRPC = NULL;
+    }
+    if(ctx->grpc.hDll) {
+        FreeLibrary(ctx->grpc.hDll);
+    }
+    ZeroMemory(&ctx->grpc, sizeof(ctx->grpc));
+    // Close the MR-RPC connection:
+#ifdef _WIN32
     if(ctx->hRPC) { 
         RpcBindingFree(ctx->hRPC);
         ctx->hRPC = NULL;
@@ -222,6 +200,7 @@ VOID LeechRPC_RpcClose(PLEECHRPC_CLIENT_CONTEXT ctx)
         RpcStringFreeA(&ctx->szStringBinding);
         ctx->szStringBinding = NULL;
     }
+#endif /* _WIN32 */
 }
 
 VOID LeechRPC_Close(_Inout_ PLC_CONTEXT ctxLC)
@@ -238,46 +217,174 @@ VOID LeechRPC_Close(_Inout_ PLC_CONTEXT ctxLC)
     while(ctx->fHousekeeperThreadIsRunning) {
         SwitchToThread();
     }
-    if(ctx->hPipeMem_Rd) {
-        CloseHandle(ctx->hPipeMem_Rd);
-        ctx->hPipeMem_Rd = NULL;
-    }
-    if(ctx->hPipeMem_Wr) {
-        CloseHandle(ctx->hPipeMem_Wr);
-        ctx->hPipeMem_Wr = NULL;
-    }
     LeechRPC_RpcClose(ctx);
     LeechRPC_CompressClose(&ctx->Compress);
+    if(ctx->hHousekeeperThread) { CloseHandle(ctx->hHousekeeperThread); }
     LocalFree(ctx);
     ctxLC->hDevice = 0;
 }
 
+#ifdef _WIN32
+
+/*
+* Helper function - connect with custom ntlm credentials to target:
+*/
+_Success_(return)
+BOOL LeechRPC_RpcInitialize_NtlmWithUserCreds(_In_ PLC_CONTEXT ctxLC, _In_ PLEECHRPC_CLIENT_CONTEXT ctx)
+{
+    BOOL fResult = FALSE;
+    RPC_STATUS status;
+    WCHAR wszTcpAddr[MAX_PATH];
+    WCHAR wszAuthIdentityUser[MAX_PATH] = { 0 };
+    WCHAR wszAuthIdentityDomain[MAX_PATH] = { 0 };
+    WCHAR wszAuthIdentityPassword[MAX_PATH] = { 0 };
+    LPVOID pvAuthBuffer = NULL;
+    ULONG cbAuthBuffer = 0;
+    ULONG AuthenticationPackage = 0;
+    HANDLE LsaHandle = NULL;
+    SEC_WINNT_AUTH_IDENTITY_W AuthIdentity = { 0 };
+    RPC_SECURITY_QOS RpcSecurityQOS = { 0 };
+    LSA_STRING PackageName = { 0 };
+    CREDUI_INFOW credui = { 0 };
+    RpcSecurityQOS.Version = RPC_C_SECURITY_QOS_VERSION;
+    RpcSecurityQOS.Capabilities = RPC_C_QOS_CAPABILITIES_DEFAULT;
+    RpcSecurityQOS.IdentityTracking = RPC_C_QOS_IDENTITY_DYNAMIC;
+    RpcSecurityQOS.ImpersonationType = RPC_C_IMP_LEVEL_IDENTIFY;
+    // prepare ui prompt:
+    swprintf_s(wszTcpAddr, _countof(wszTcpAddr), L"Enter your credentials to connect to: %S", ctx->szTcpAddr);
+    credui.cbSize = sizeof(credui);
+    credui.hwndParent = NULL;
+    credui.pszMessageText = wszTcpAddr;
+    credui.pszCaptionText = L"Enter network credentials";
+    credui.hbmBanner = NULL;
+    // get lsa package:
+    if(ERROR_SUCCESS != LsaConnectUntrusted(&LsaHandle)) { goto fail; }
+    PackageName.Buffer = MICROSOFT_KERBEROS_NAME_A;
+    PackageName.Length = (USHORT)strlen(PackageName.Buffer);
+    PackageName.MaximumLength = (USHORT)strlen(PackageName.Buffer);
+    if(ERROR_SUCCESS != LsaLookupAuthenticationPackage(LsaHandle, &PackageName, &AuthenticationPackage)) { goto fail; }
+    // get user creds via credprompt (unless user already set both user & password by command line):
+    AuthIdentity.Domain = wszAuthIdentityDomain;
+    AuthIdentity.DomainLength = (DWORD)_countof(wszAuthIdentityDomain);
+    AuthIdentity.User = wszAuthIdentityUser;
+    AuthIdentity.UserLength = (DWORD)_countof(wszAuthIdentityUser);
+    AuthIdentity.Password = wszAuthIdentityPassword;
+    AuthIdentity.PasswordLength = (DWORD)_countof(wszAuthIdentityPassword);
+    AuthIdentity.Flags = SEC_WINNT_AUTH_IDENTITY_UNICODE;
+    if(ctx->szAuthNtlmUserInitOnly && ctx->szAuthNtlmPasswordInitOnly) {
+        AuthIdentity.DomainLength = 0;
+    } else {
+        if(ERROR_SUCCESS != CredUIPromptForWindowsCredentialsW(&credui, 0, &AuthenticationPackage, NULL, 0, &pvAuthBuffer, &cbAuthBuffer, NULL, CREDUIWIN_GENERIC)) { goto fail; }
+        // unpack user creds:
+        if(FALSE == CredUnPackAuthenticationBufferW(CRED_PACK_PROTECTED_CREDENTIALS, pvAuthBuffer, cbAuthBuffer, AuthIdentity.User, &AuthIdentity.UserLength, AuthIdentity.Domain, &AuthIdentity.DomainLength, AuthIdentity.Password, &AuthIdentity.PasswordLength)) { goto fail; }
+        if(AuthIdentity.UserLength && (wszAuthIdentityUser[AuthIdentity.UserLength - 1] == 0)) { AuthIdentity.UserLength--; }
+        if(AuthIdentity.DomainLength && (wszAuthIdentityDomain[AuthIdentity.DomainLength - 1] == 0)) { AuthIdentity.DomainLength--; }
+        if(AuthIdentity.PasswordLength && (wszAuthIdentityPassword[AuthIdentity.PasswordLength - 1] == 0)) { AuthIdentity.PasswordLength--; }
+    }
+    if(ctx->szAuthNtlmUserInitOnly) {
+        AuthIdentity.UserLength = _snwprintf_s(wszAuthIdentityUser, _countof(wszAuthIdentityUser), _TRUNCATE, L"%S", ctx->szAuthNtlmUserInitOnly);
+    }
+    if(ctx->szAuthNtlmPasswordInitOnly) {
+        AuthIdentity.PasswordLength = _snwprintf_s(wszAuthIdentityPassword, _countof(wszAuthIdentityPassword), _TRUNCATE, L"%S", ctx->szAuthNtlmPasswordInitOnly);
+    }
+    status = RpcBindingSetAuthInfoExW(
+        ctx->hRPC,
+        NULL,
+        RPC_C_AUTHN_LEVEL_PKT_PRIVACY,
+        RPC_C_AUTHN_WINNT,
+        &AuthIdentity,
+        RPC_C_AUTHZ_DEFAULT,
+        &RpcSecurityQOS);
+    if(status) {
+        lcprintf(ctxLC, "REMOTE: Failed to set connection security for connection, Error code: 0x%08x\n", status);
+        LeechRPC_RpcClose(ctx);
+        goto fail;
+    }
+    fResult = TRUE;
+fail:
+    SecureZeroMemory(pvAuthBuffer, cbAuthBuffer);
+    SecureZeroMemory(wszAuthIdentityPassword, sizeof(wszAuthIdentityPassword));
+    if(LsaHandle) { LsaDeregisterLogonProcess(LsaHandle); LsaHandle = NULL; }
+    if(pvAuthBuffer) { CoTaskMemFree(pvAuthBuffer); pvAuthBuffer = NULL; }
+    return fResult;
+}
+
 _Success_(return)
 BOOL LeechRPC_RpcInitialize(_In_ PLC_CONTEXT ctxLC, _In_ PLEECHRPC_CLIENT_CONTEXT ctx)
 {
+    LPSTR szTcpAddr;
     RPC_STATUS status;
     RPC_SECURITY_QOS RpcSecurityQOS = { 0 };
     LeechRPC_RpcClose(ctx);
-    status = RpcStringBindingComposeA(
-        CLSID_BINDING_INTERFACE_LEECHRPC,
-        "ncacn_ip_tcp",
-        ctx->szTcpAddr,
-        ctx->szTcpPort,
-        NULL,
-        &ctx->szStringBinding);
-    if(status) {
-        lcprintf(ctxLC, "RPC: Failed compose binding: Error code: 0x%08x\n", status);
-        LeechRPC_RpcClose(ctx);
-        return FALSE;
+    // compose binding string:
+    if(ctx->fIsProtoSmb) {
+        if((ctx->szTcpAddr[0] == 0) || !_stricmp("localhost", ctx->szTcpAddr) || !_stricmp("127.0.0.1", ctx->szTcpAddr)) {
+            szTcpAddr = NULL;
+        } else {
+            szTcpAddr = ctx->szTcpAddr;
+        }
+        status = RpcStringBindingComposeA(
+            CLSID_BINDING_INTERFACE_LEECHRPC,
+            "ncacn_np",
+            szTcpAddr,
+            "\\pipe\\LeechAgent",
+            NULL,
+            &ctx->szStringBinding);
+        if(status) {
+            lcprintf(ctxLC, "REMOTE: Failed compose binding: Error code: 0x%08x\n", status);
+            LeechRPC_RpcClose(ctx);
+            return FALSE;
+        }
     }
+    if(ctx->fIsProtoRpc) {
+        status = RpcStringBindingComposeA(
+            CLSID_BINDING_INTERFACE_LEECHRPC,
+            "ncacn_ip_tcp",
+            ctx->szTcpAddr,
+            ctx->szTcpPort,
+            NULL,
+            &ctx->szStringBinding);
+        if(status) {
+            lcprintf(ctxLC, "REMOTE: Failed compose binding: Error code: 0x%08x\n", status);
+            LeechRPC_RpcClose(ctx);
+            return FALSE;
+        }
+    }
+    // create binding:
     status = RpcBindingFromStringBindingA(ctx->szStringBinding, &ctx->hRPC);
     if(status) {
-        lcprintf(ctxLC, "RPC: Failed create binding: Error code: 0x%08x\n", status);
+        lcprintf(ctxLC, "REMOTE: Failed create binding: Error code: 0x%08x\n", status);
         LeechRPC_RpcClose(ctx);
         return FALSE;
     }
-    if(!ctx->fAllowInsecure) {
-        RpcSecurityQOS.Version = 1;
+    // set connection security (if any):
+    if(ctx->fIsAuthNTLM) {
+        if(ctx->fIsAuthNTLMCredPrompt) {
+            if(!LeechRPC_RpcInitialize_NtlmWithUserCreds(ctxLC, ctx)) { return FALSE; }
+        } else {
+            // NTLM - use default credentials (current user) or user-supplied credentials via prompt:
+            RpcSecurityQOS.Version = RPC_C_SECURITY_QOS_VERSION;
+            RpcSecurityQOS.Capabilities = RPC_C_QOS_CAPABILITIES_DEFAULT;
+            RpcSecurityQOS.IdentityTracking = RPC_C_QOS_IDENTITY_DYNAMIC;
+            RpcSecurityQOS.ImpersonationType = RPC_C_IMP_LEVEL_IDENTIFY;
+            status = RpcBindingSetAuthInfoExA(
+                ctx->hRPC,
+                NULL,
+                RPC_C_AUTHN_LEVEL_PKT_PRIVACY,
+                RPC_C_AUTHN_WINNT,
+                NULL,
+                RPC_C_AUTHZ_DEFAULT,
+                &RpcSecurityQOS);
+            if(status) {
+                lcprintf(ctxLC, "REMOTE: Failed to set connection security for connection, Error code: 0x%08x\n", status);
+                LeechRPC_RpcClose(ctx);
+                return FALSE;
+            }
+        }
+    }
+    if(ctx->fIsAuthKerberos) {
+        // Kerberos - use specified SPN for mutual authentication:
+        RpcSecurityQOS.Version = RPC_C_SECURITY_QOS_VERSION;
         RpcSecurityQOS.Capabilities = RPC_C_QOS_CAPABILITIES_MUTUAL_AUTH;
         RpcSecurityQOS.IdentityTracking = RPC_C_QOS_IDENTITY_DYNAMIC;
         RpcSecurityQOS.ImpersonationType = RPC_C_IMP_LEVEL_IDENTIFY;
@@ -287,17 +394,63 @@ BOOL LeechRPC_RpcInitialize(_In_ PLC_CONTEXT ctxLC, _In_ PLEECHRPC_CLIENT_CONTEX
             RPC_C_AUTHN_LEVEL_PKT_PRIVACY,
             RPC_C_AUTHN_GSS_KERBEROS,
             NULL,
-            0,
+            RPC_C_AUTHZ_NONE,
             &RpcSecurityQOS);
         if(status) {
-            lcprintf(ctxLC, "RPC: Failed to set connection security: SPN: '%s', Error code: 0x%08x\n", ctx->szRemoteSPN, status);
-            lcprintf(ctxLC, "     Maybe try kerberos security disable by specify SPN 'insecure' if server allows...\n");
+            lcprintf(ctxLC, "REMOTE: Failed to set connection security: SPN: '%s', Error code: 0x%08x\n", ctx->szRemoteSPN, status);
+            lcprintf(ctxLC, "        Maybe try kerberos security disable by specify SPN 'insecure' if server allows...\n");
             LeechRPC_RpcClose(ctx);
             return FALSE;
         }
     }
     lcprintfv_fn(ctxLC, "'%s'\n", ctx->szStringBinding);
-    ctx->fIsRpc = TRUE;
+    return TRUE;
+}
+
+#endif /* _WIN32 */
+#if defined(LINUX) || defined(MACOS)
+
+_Success_(return)
+BOOL LeechRPC_RpcInitialize(_In_ PLC_CONTEXT ctxLC, _In_ PLEECHRPC_CLIENT_CONTEXT ctx)
+{
+    lcprintf(ctxLC, "REMOTE: MS-RPC with '-remote rpc' or '-remote smb' is only supported on Windows. gRPC is supported.\n");
+    return FALSE;
+}
+
+#endif /* LINUX || MACOS */
+
+_Success_(return)
+BOOL LeechRPC_GRpcInitialize(_In_ PLC_CONTEXT ctxLC, _In_ PLEECHRPC_CLIENT_CONTEXT ctx)
+{
+    DWORD dwTcpPort = strtol(ctx->szTcpPort, NULL, 10);
+    ctx->grpc.hDll = LoadLibraryA("libleechgrpc"LC_LIBRARY_FILETYPE);
+    if(!ctx->grpc.hDll) {
+        lcprintf(ctxLC, "REMOTE: Failed to load 'libleechgrpc"LC_LIBRARY_FILETYPE"'\n");
+        return FALSE;
+    }
+    ctx->grpc.pfn_leechgrpc_client_create_insecure = (pfn_leechgrpc_client_create_insecure)GetProcAddress(ctx->grpc.hDll, "leechgrpc_client_create_insecure");
+    ctx->grpc.pfn_leechgrpc_client_create_secure_p12 = (pfn_leechgrpc_client_create_secure_p12)GetProcAddress(ctx->grpc.hDll, "leechgrpc_client_create_secure_p12");
+    ctx->grpc.pfn_leechgrpc_client_free = (pfn_leechgrpc_client_free)GetProcAddress(ctx->grpc.hDll, "leechgrpc_client_free");
+    ctx->grpc.pfn_leechgrpc_client_submit_command = (pfn_leechgrpc_client_submit_command)GetProcAddress(ctx->grpc.hDll, "leechgrpc_client_submit_command");
+    if(!ctx->grpc.pfn_leechgrpc_client_create_insecure || !ctx->grpc.pfn_leechgrpc_client_create_secure_p12 || !ctx->grpc.pfn_leechgrpc_client_free || !ctx->grpc.pfn_leechgrpc_client_submit_command) {
+        lcprintf(ctxLC, "REMOTE: Failed to load functions from 'libleechgrpc"LC_LIBRARY_FILETYPE"'\n");
+        return FALSE;
+    }
+    if(ctx->fIsAuthInsecure) {
+        ctx->grpc.hGRPC = ctx->grpc.pfn_leechgrpc_client_create_insecure(ctx->szTcpAddr, dwTcpPort);
+    } else {
+        ctx->grpc.hGRPC = ctx->grpc.pfn_leechgrpc_client_create_secure_p12(
+            ctx->szTcpAddr,
+            dwTcpPort,
+            ctx->grpc.szServerCertHostnameOverride[0] ? ctx->grpc.szServerCertHostnameOverride : NULL,
+            ctx->grpc.szServerCertCaPath[0] ? ctx->grpc.szServerCertCaPath : NULL,
+            ctx->grpc.szClientTlsP12Path,
+            ctx->grpc.szClientTlsP12Password);
+    }
+    if(!ctx->grpc.hGRPC) {
+        lcprintf(ctxLC, "REMOTE: Failed to create gRPC client connection\n");
+        return FALSE;
+    }
     return TRUE;
 }
 
@@ -532,6 +685,7 @@ BOOL LeechRPC_Command(
     PLEECHRPC_MSG_BIN pMsgRsp = NULL;
     // 1: prepare message to send
     if(!pbDataIn && cbDataIn) { return FALSE; }
+    if(fCMD & 0x2000000000000000) { return FALSE; }     // command is marked as no-remote.
     if(!(pMsgReq = LocalAlloc(0, sizeof(LEECHRPC_MSG_BIN) + cbDataIn))) { return FALSE; }
     ZeroMemory(pMsgReq, sizeof(LEECHRPC_MSG_BIN));
     pMsgReq->tpMsg = LEECHRPC_MSGTYPE_COMMAND_REQ;
@@ -549,7 +703,7 @@ BOOL LeechRPC_Command(
     if(result) {
         if(pcbDataOut) { *pcbDataOut = pMsgRsp->cb; }
         if(ppbDataOut) {
-            if(*ppbDataOut = LocalAlloc(0, pMsgRsp->cb)) {
+            if((*ppbDataOut = LocalAlloc(0, pMsgRsp->cb))) {
                 memcpy(*ppbDataOut, pMsgRsp->pb, pMsgRsp->cb);
             } else {
                 result = FALSE;
@@ -571,74 +725,136 @@ BOOL LeechRPC_Command(
 _Success_(return)
 BOOL LeechRpc_Open(_Inout_ PLC_CONTEXT ctxLC, _Out_opt_ PPLC_CONFIG_ERRORINFO ppLcCreateErrorInfo)
 {
+    BOOL f;
     PLEECHRPC_CLIENT_CONTEXT ctx;
     CHAR _szBufferArg[MAX_PATH], _szBufferOpt[MAX_PATH];
     LEECHRPC_MSG_OPEN MsgReq = { 0 };
     PLEECHRPC_MSG_OPEN pMsgRsp = NULL;
     LPSTR szArg1, szArg2, szArg3;
-    LPSTR szOpt[3];
+    LPSTR aszOpt[6];
     DWORD i, dwPort = 0;
-    HANDLE hThread;
     int(*pfn_printf_opt_tmp)(_In_z_ _Printf_format_string_ char const* const _Format, ...);
     if(ppLcCreateErrorInfo) { *ppLcCreateErrorInfo = NULL; }
     ctx = (PLEECHRPC_CLIENT_CONTEXT)LocalAlloc(LMEM_ZEROINIT, sizeof(LEECHRPC_CLIENT_CONTEXT));
     if(!ctx) { return FALSE; }
     ctxLC->hDevice = (HANDLE)ctx;
-    if(!_stricmp(ctxLC->Config.szDeviceName, "rpc")) {
+    if(!_stricmp(ctxLC->Config.szDeviceName, "grpc")) { ctx->fIsProtoGRpc = TRUE; }
+    if(!_stricmp(ctxLC->Config.szDeviceName, "rpc")) { ctx->fIsProtoRpc = TRUE; }
+    if(!_stricmp(ctxLC->Config.szDeviceName, "smb")) { ctx->fIsProtoSmb = TRUE; }
+    if(!ctx->fIsProtoGRpc && !ctx->fIsProtoRpc && !ctx->fIsProtoSmb) {
+        lcprintf(ctxLC, "REMOTE: ERROR: No valid remote transport protocol specified.\n");
+        goto fail;
+    }
+    if(ctx->fIsProtoRpc || ctx->fIsProtoSmb) {
         // RPC SPECIFIC INITIALIZATION BELOW:
         ctxLC->Rpc.fCompress = !ctxLC->Config.fRemoteDisableCompress;
         // parse arguments
         Util_Split3(ctxLC->Config.szRemote + 6, ':', _szBufferArg, &szArg1, &szArg2, &szArg3);
         if(!szArg1 || !szArg1[0] || !szArg2 || !szArg2[0]) { goto fail; }
-        // Argument1 : Kerberos SPN or "insecure".
+        // Argument1 : Auth method, insecure | ntlm | kerberos_spn
         if(!_stricmp("insecure", szArg1)) {
-            ctx->fAllowInsecure = TRUE;
+            ctx->fIsAuthInsecure = TRUE;
+        } else if(!_stricmp("ntlm", szArg1)) {
+            ctx->fIsAuthNTLM = TRUE;
         } else {
             strncpy_s(ctx->szRemoteSPN, _countof(ctx->szRemoteSPN), szArg1, MAX_PATH);
+            ctx->fIsAuthKerberos = TRUE;
         }
         // Argument2 : Tcp Address.
         strncpy_s(ctx->szTcpAddr, _countof(ctx->szTcpAddr), szArg2, MAX_PATH);
         // Argument3 : Options.
         if(szArg3[0]) {
-            Util_Split3(szArg3, ',', _szBufferOpt, &szOpt[0], &szOpt[1], &szOpt[2]);
-            for(i = 0; i < 3; i++) {
-                if(0 == _stricmp("nocompress", szOpt[i])) {
+            Util_SplitN(szArg3, ',', 5, _szBufferOpt, aszOpt);
+            for(i = 0; i < 5; i++) {
+                if(0 == _stricmp("nocompress", aszOpt[i])) {
                     ctxLC->Rpc.fCompress = FALSE;
                 }
-                if(0 == _strnicmp("port=", szOpt[i], 5)) {
-                    dwPort = atoi(szOpt[i] + 5);
+                if(0 == _strnicmp("port=", aszOpt[i], 5)) {
+                    dwPort = atoi(aszOpt[i] + 5);
+                }
+                if(0 == _stricmp("logon", aszOpt[i])) {
+                    ctx->fIsAuthNTLMCredPrompt = ctx->fIsAuthNTLM;
+                }
+                if(0 == _strnicmp("user=", aszOpt[i], 5)) {
+                    ctx->szAuthNtlmUserInitOnly = aszOpt[i] + 5;
+                }
+                if(0 == _strnicmp("password=", aszOpt[i], 9)) {
+                    ctx->szAuthNtlmPasswordInitOnly = aszOpt[i] + 9;
                 }
             }
         }
+        ctx->fIsAuthNTLMCredPrompt = ctx->fIsAuthNTLMCredPrompt || ctx->szAuthNtlmUserInitOnly || ctx->szAuthNtlmPasswordInitOnly;
         if(dwPort == 0) {
             dwPort = 28473; // default port
         }
         _itoa_s(dwPort, ctx->szTcpPort, 6, 10);
         // initialize rpc connection and ping
-        if(!LeechRPC_RpcInitialize(ctxLC, ctx)) {
-            lcprintf(ctxLC, "RPC: ERROR: Unable to connect to remote service '%s'\n", ctxLC->Config.szRemote);
+        f = LeechRPC_RpcInitialize(ctxLC, ctx);
+        ctx->szAuthNtlmUserInitOnly = NULL;
+        ctx->szAuthNtlmPasswordInitOnly = NULL;
+        SecureZeroMemory(_szBufferOpt, sizeof(_szBufferOpt));
+        if(!f) {
+            lcprintf(ctxLC, "REMOTE: ERROR: Unable to connect to remote service '%s'\n", ctxLC->Config.szRemote);
             goto fail;
         }
         if(!LeechRPC_Ping(ctxLC)) {
-            lcprintf(ctxLC, "RPC: ERROR: Unable to ping remote service '%s'\n", ctxLC->Config.szRemote);
+            lcprintf(ctxLC, "REMOTE: ERROR: Unable to ping remote service '%s'\n", ctxLC->Config.szRemote);
             goto fail;
         }
-    } else if(!_stricmp(ctxLC->Config.szDeviceName, "pipe")) {
-        // PIPE SPECIFIC INITIALIZATION BELOW:
+    }
+    if(ctx->fIsProtoGRpc) {
+        // RPC SPECIFIC INITIALIZATION BELOW:
+        ctxLC->Rpc.fCompress = !ctxLC->Config.fRemoteDisableCompress;
         // parse arguments
-        Util_Split2(ctxLC->Config.szRemote + 7, ':', _szBufferArg, &szArg1, &szArg2);
-        if(!szArg1 || !szArg2) { goto fail; }
-        ctx->hPipeMem_Rd = (HANDLE)_atoi64(szArg1);
-        ctx->hPipeMem_Wr = (HANDLE)_atoi64(szArg2);
-        if(!ctx->hPipeMem_Rd || !ctx->hPipeMem_Wr) { goto fail; }
-        // ping parent process via the pipe
+        Util_Split3(ctxLC->Config.szRemote + 7, ':', _szBufferArg, &szArg1, &szArg2, &szArg3);
+        if(!szArg1 || !szArg2 || !szArg2[0]) { goto fail; }
+        // Argument1 : Auth method, insecure
+        if(!_stricmp("insecure", szArg1)) {
+            ctx->fIsAuthInsecure = TRUE;
+        } else if(szArg1[0]) {
+            strncpy_s(ctx->grpc.szServerCertHostnameOverride, _countof(ctx->grpc.szServerCertHostnameOverride), szArg1, _TRUNCATE);
+        }
+        // Argument2 : Tcp Address.
+        strncpy_s(ctx->szTcpAddr, _countof(ctx->szTcpAddr), szArg2, MAX_PATH);
+        // Argument3 : Options.
+        if(szArg3[0]) {
+            Util_SplitN(szArg3, ',', 6, _szBufferOpt, aszOpt);
+            for(i = 0; i < 6; i++) {
+                if(0 == _stricmp("nocompress", aszOpt[i])) {
+                    ctxLC->Rpc.fCompress = FALSE;
+                }
+                if(0 == _strnicmp("port=", aszOpt[i], 5)) {
+                    dwPort = atoi(aszOpt[i] + 5);
+                }
+                if(0 == _strnicmp("client-cert-p12-password=", aszOpt[i], 25)) {
+                    strncpy_s(ctx->grpc.szClientTlsP12Password, _countof(ctx->grpc.szClientTlsP12Password), aszOpt[i] + 25, _TRUNCATE);
+                }
+                if(0 == _strnicmp("client-cert-p12=", aszOpt[i], 16)) {
+                    strncpy_s(ctx->grpc.szClientTlsP12Path, _countof(ctx->grpc.szClientTlsP12Path), aszOpt[i] + 16, _TRUNCATE);
+                }
+                if(0 == _strnicmp("server-cert=", aszOpt[i], 12)) {
+                    strncpy_s(ctx->grpc.szServerCertCaPath, _countof(ctx->grpc.szServerCertCaPath), aszOpt[i] + 12, _TRUNCATE);
+                }
+                if(0 == _strnicmp("server-cert-host-override=", aszOpt[i], 26)) {
+                    strncpy_s(ctx->grpc.szServerCertHostnameOverride, _countof(ctx->grpc.szServerCertHostnameOverride), aszOpt[i] + 26, _TRUNCATE);
+                }
+            }
+        }
+        if(dwPort == 0) {
+            dwPort = 28474; // default port
+        }
+        _itoa_s(dwPort, ctx->szTcpPort, 6, 10);
+        // initialize rpc connection and ping
+        f = LeechRPC_GRpcInitialize(ctxLC, ctx);
+        SecureZeroMemory(_szBufferOpt, sizeof(_szBufferOpt));
+        if(!f) {
+            lcprintf(ctxLC, "REMOTE: ERROR: Unable to connect to remote gRPC service '%s'\n", ctxLC->Config.szRemote);
+            goto fail;
+        }
         if(!LeechRPC_Ping(ctxLC)) {
-            lcprintf(ctxLC, "PIPE: ERROR: Unable to ping remote service '%s'\n", ctxLC->Config.szRemote);
+            lcprintf(ctxLC, "REMOTE: ERROR: Unable to ping remote gRPC service '%s'\n", ctxLC->Config.szRemote);
             goto fail;
         }
-        ctxLC->Rpc.fCompress = FALSE;       // compress = default off on pipe
-    } else {
-        goto fail;
     }
     if(0 == _strnicmp(ctxLC->Config.szDevice, "existingremote", 14)) {
         for(i = 14; i < _countof(ctxLC->Config.szDevice); i++) {
@@ -656,7 +872,7 @@ BOOL LeechRpc_Open(_Inout_ PLC_CONTEXT ctxLC, _Out_opt_ PPLC_CONFIG_ERRORINFO pp
     ZeroMemory(MsgReq.cfg.szRemote, _countof(MsgReq.cfg.szRemote));
     MsgReq.cfg.pfn_printf_opt = 0;
     if(!LeechRPC_SubmitCommand(ctxLC, (PLEECHRPC_MSG_HDR)&MsgReq, LEECHRPC_MSGTYPE_OPEN_RSP, (PPLEECHRPC_MSG_HDR)&pMsgRsp)) {
-        lcprintf(ctxLC, "RPC: ERROR: Unable to open remote device #1 '%s'\n", ctxLC->Config.szDevice);
+        lcprintf(ctxLC, "REMOTE: ERROR: Unable to open remote device #1 '%s'\n", ctxLC->Config.szDevice);
         goto fail;
     }
     if(!pMsgRsp->fValidOpen) {
@@ -666,35 +882,34 @@ BOOL LeechRpc_Open(_Inout_ PLC_CONTEXT ctxLC, _Out_opt_ PPLC_CONFIG_ERRORINFO pp
                 memcpy(*ppLcCreateErrorInfo, &pMsgRsp->errorinfo, pMsgRsp->errorinfo.cbStruct);
             }
         }
-        lcprintf(ctxLC, "RPC: ERROR: Unable to open remote device #2 '%s'\n", ctxLC->Config.szDevice);
+        lcprintf(ctxLC, "REMOTE: ERROR: Unable to open remote device #2 '%s'\n", ctxLC->Config.szDevice);
         goto fail;
     }
     // sanity check positive result from remote service
     if(pMsgRsp->cfg.dwVersion != LC_CONFIG_VERSION) {
-        lcprintf(ctxLC, "RPC: ERROR: Invalid message received from remote service.\n");
+        lcprintf(ctxLC, "REMOTE: ERROR: Invalid message received from remote service.\n");
         goto fail;
     }
     if(ctxLC->Rpc.fCompress && pMsgRsp->cfg.fRemoteDisableCompress) {
         ctxLC->Config.fRemoteDisableCompress = TRUE;
         ctxLC->Rpc.fCompress = FALSE;
-        lcprintfv(ctxLC, "RPC: INFO: Compression disabled.\n");
+        lcprintfv(ctxLC, "REMOTE: INFO: Compression disabled.\n");
     }
     // all ok - initialize this rpc device stub.
-    hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)LeechRPC_KeepaliveThreadClient, ctxLC, 0, NULL);
-    if(hThread) { CloseHandle(hThread); hThread = 0; }
+    ctx->hHousekeeperThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)LeechRPC_KeepaliveThreadClient, ctxLC, 0, NULL);
     strncpy_s(pMsgRsp->cfg.szRemote, sizeof(pMsgRsp->cfg.szRemote), ctxLC->Config.szRemote, _TRUNCATE); // ctx from remote doesn't contain remote info ...
     pfn_printf_opt_tmp = ctxLC->Config.pfn_printf_opt;
     memcpy(&ctxLC->Config, &pMsgRsp->cfg, sizeof(LC_CONFIG));
     ctxLC->Config.pfn_printf_opt = pfn_printf_opt_tmp;
     ctxLC->Config.fRemote = TRUE;
-    ctxLC->fMultiThread = ctx->fIsRpc;        // RPC = multi-thread, PIPE = single-thread access
+    ctxLC->fMultiThread = TRUE;
     ctxLC->pfnClose = LeechRPC_Close;
     ctxLC->pfnReadScatter = LeechRPC_ReadScatter;
     ctxLC->pfnWriteScatter = LeechRPC_WriteScatter;
     ctxLC->pfnGetOption = LeechRPC_GetOption;
     ctxLC->pfnSetOption = LeechRPC_SetOption;
     ctxLC->pfnCommand = LeechRPC_Command;
-    lcprintfv(ctxLC, "RPC: Successfully opened remote device: %s\n", ctxLC->Config.szDeviceName);
+    lcprintfv(ctxLC, "REMOTE: Successfully opened remote device: %s\n", ctxLC->Config.szDeviceName);
     LocalFree(pMsgRsp);
     return TRUE;
 fail:
@@ -702,15 +917,3 @@ fail:
     LocalFree(pMsgRsp);
     return FALSE;
 }
-
-#endif /* _WIN32 */
-#ifdef LINUX
-
-_Success_(return)
-BOOL LeechRpc_Open(_Inout_ PLC_CONTEXT ctxLC, _Out_opt_ PPLC_CONFIG_ERRORINFO ppLcCreateErrorInfo)
-{
-    if(ppLcCreateErrorInfo) { *ppLcCreateErrorInfo = NULL; }
-    return FALSE;
-}
-
-#endif /* LINUX */
